/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "gldraw.h"

#include "constants.h"

#include "glocke.h"
#include "rabbit.h"
#include "renderedtext.h"
#include "qwtype.h"

#include <math.h>
#include <algorithm>

OpenGLDrawContext::OpenGLDrawContext(Game *game)
    : m_game(game),
      m_rotated(false),
      m_width(0),
      m_height(0),
      m_scale(1.)
{
}

OpenGLDrawContext::~OpenGLDrawContext()
{
}

void
OpenGLDrawContext::initializeGL()
{
    GLocke::init();
    rabbitSetupLighting();
}

void
OpenGLDrawContext::resizeGL(int width, int height)
{
    m_width = width;
    m_height = height;
    m_rotated = GLocke::resize(width, height);

    m_scale = 1.;
    if (width > Constants::WINDOW_WIDTH) {
        m_scale = (float)width / (float)Constants::WINDOW_WIDTH;
    }

    m_game->resize(width/m_scale, height/m_scale);
}

void
OpenGLDrawContext::paintGL()
{
    m_game->render(this);
}


void
OpenGLDrawContext::mapCoords(int &x, int &y)
{
    if (m_rotated) {
        std::swap(x, y);
        x = m_height - x;
    }
    x /= m_scale;
    y /= m_scale;
}

void
OpenGLDrawContext::setColor(enum Color color, float opacity)
{
    float r, g, b, a=opacity;
    RGBColor::map(color, r, g, b);
    GLocke::color(r, g, b, a);
}

void
OpenGLDrawContext::drawTriangle(float r, float g, float b, float a,
        int x1, int y1, int x2, int y2, int x3, int y3)
{
    float triangle_vertices[] = {
        x1, y1, .5, r, g, b, a*.9, // For a slight, fancy effect
        x2, y2, .5, r, g, b, a,
        x3, y3, .5, r, g, b, a,
    };

    int stride = sizeof(float) * (3 + 4);
    int elements = sizeof(triangle_vertices) / stride;

    GLocke::applyModelView(3, sizeof(float) * (3 + 4), triangle_vertices, elements);

    GLocke::draw2d(3, triangle_vertices);
}

void
OpenGLDrawContext::drawBrick(enum Color color, int x, int y, float rotation, float opacity, bool empty, float darken)
{
    int ir, ig, ib;
    RGBColor::map(color, ir, ig, ib);
    float r = darken * ir / 255.f,
          g = darken * ig / 255.f,
          b = darken * ib / 255.f,
          a = opacity;

    qwMatrix m = GLocke::modelviewMatrix();
    if (empty) {
        a *= 0.2 + 0.5 * float(y)/float(m_height);
        m *= qwMatrix::translation(0.f, 0.f, -.2f);
        m *= qwMatrix::scaling(1.f, 1.f, .1f);
    }
    m *= qwMatrix::translation(x+Constants::BRICK_WIDTH*.5, y+Constants::BRICK_HEIGHT*.5, 0.);
    if (empty) {
        m *= qwMatrix::scaling(Constants::BRICK_WIDTH*0.95, Constants::BRICK_HEIGHT*0.95, .3);
    } else {
        m *= qwMatrix::scaling(Constants::BRICK_WIDTH, Constants::BRICK_HEIGHT, .3);
    }
    m *= qwMatrix::rotation(rotation, .25, .5, 1);

    /**
     * Cube vertex data might be batched up for rendering, therefore we must
     * assume that vertex data is concatenated. As we use triangle strip
     * rendering, we must degenerate the triangles in-between when "jumping"
     * to the left and right side of the cube as well as between cubes. We
     * therefore duplicate the first and last vertex, so that when rendering
     * two cubes in one draw call, the triangle in between will be degenerated.
     **/

    float light = 1.2, dark = 0.9;
    float rl = r * light, gl = g * light, bl = b * light;
    float rd = r * dark, gd = g * dark, bd = b * dark;

    float cube_vertices[] = {
        // Degenerated triangle (first of this one, so we can concatenate vertex data)
        -0.5, -0.5, +0.5, r, g, b, a,

        // Front side
        -0.5, -0.5, +0.5, rl, gl, bl, a,
        +0.5, -0.5, +0.5, r, g, b, a,
        -0.5, +0.5, +0.5, r, g, b, a,
        +0.5, +0.5, +0.5, r, g, b, a,

        // Bottom side
        -0.5, +0.5, -0.5, r, g, b, a,
        +0.5, +0.5, -0.5, r, g, b, a,

        // Back side
        -0.5, -0.5, -0.5, r, g, b, a,
        +0.5, -0.5, -0.5, r, g, b, a,

        // Top side
        -0.5, -0.5, +0.5, rd, gd, bd, a,
        +0.5, -0.5, +0.5, rd, gd, bd, a,

        // Degenerated triangle (last of previous + first of next)
        +0.5, -0.5, +0.5, r, g, b, a,
        -0.5, -0.5, -0.5, r, g, b, a,

        // Left side
        -0.5, -0.5, -0.5, r, g, b, a,
        -0.5, -0.5, +0.5, r, g, b, a,
        -0.5, +0.5, -0.5, r, g, b, a,
        -0.5, +0.5, +0.5, r, g, b, a,

        // Degenerated triangle (last of previous + first of next)
        -0.5, +0.5, +0.5, r, g, b, a,
        +0.5, -0.5, -0.5, r, g, b, a,

        // Right side
        +0.5, -0.5, -0.5, r, g, b, a,
        +0.5, -0.5, +0.5, r, g, b, a,
        +0.5, +0.5, -0.5, r, g, b, a,
        +0.5, +0.5, +0.5, r, g, b, a,

        // Degenerated triangle (last of this one, so we can concatenate vertex data)
        +0.5, +0.5, +0.5, r, g, b, a,
    };
    int components = (3 + 4);
    int elements = sizeof(cube_vertices) / (sizeof(float) * components);

    for (int i=0; i<elements; i++) {
        float *vtx = cube_vertices + i * components;
        float v[4] = { vtx[0], vtx[1], vtx[2], 1.f };
        m.apply(v);
        vtx[0] = v[0] / v[3];
        vtx[1] = v[1] / v[3];
        vtx[2] = v[2] / v[3];
    }

    GLocke::draw25d(elements, cube_vertices);
}


void
OpenGLDrawContext::draw_begin()
{
    GLocke::begin();
    GLocke::scale(m_scale, m_scale, 1.);
}

void
OpenGLDrawContext::draw_background(enum Color c1, enum Color c2)
{
    float r1, g1, b1, r2, g2, b2;
    RGBColor::map(c1, r1, g1, b1);
    RGBColor::map(c2, r2, g2, b2);

    int ticks = m_game->now();
    int size = 20;

    float i1 = .5 + cosf(ticks*.0004) * .5;
    float i2 = 1. - i1;
    float i3 = 0.15;

    float r=i3*(r1*i1+r2*i2), g=i3*(g1*i1+g2*i2), b=i3*(b1*i1+b2*i2), a=1.;
    draw_rectangle(r, g, b, a, 0, 0, m_game->m_width, m_game->m_height);
}

void
OpenGLDrawContext::draw_brick(enum Color color, int x, int y, float rotation)
{
    float time = m_game->now();
    float xx = 30.f * float(x) / float(m_width);
    float yy = 30.f * float(y) / float(m_height);
    float darken = 1.0 - 0.05 * fminf(1.0, fabsf(tanf(0.001 * time + xx + yy)));
    drawBrick(color, x, y, rotation, (color == SHADOW) ? 0.4 : 1.0, false,
            (color == SHADOW || color == WHITE) ? 1.0 : darken);
}

void
OpenGLDrawContext::draw_empty(int x, int y, bool right, bool bottom)
{
    drawBrick(BLACK, x, y, 0, 0.5, true, 1.0);
}

void
OpenGLDrawContext::draw_fading(float value, int x, int y)
{
    drawBrick(SHADOW, x, y, 0, 0.4 * value, false, 1.0);
}

void
OpenGLDrawContext::draw_score(long ticks, long score, int x, int y, int w, int h)
{
    GLocke::clearDepth();
    GLocke::pushMatrix();

    GLocke::translate(x+w/2, y+h/4, 0);
    GLocke::scale(w/2, -h/2, 1.);
    drawRabbit(ticks, score);
    GLocke::popMatrix();

}

void
OpenGLDrawContext::draw_title(int x, int y, int w, int h)
{
    GLocke::pushMatrix();

    GLocke::translate(x, y, -1.);
    GLocke::scale(100., 100., 1.);

    float time = m_game->now()*.5;
    float intensity = 1.f + fabs(cosf(time*.5)*sinf(time*.01));
    float r, g, b, a=1.f;
    RGBColor::map(COLOR_PRIMARY, r, g, b);
    GLocke::color(r*intensity, g*intensity, b*intensity, a);
    drawTitle(time);

    GLocke::popMatrix();
}

void
OpenGLDrawContext::draw_shade()
{
    // TODO: Do some funky blurring or whatever
    draw_rectangle(0.f, 0.f, 0.f, 0.9f, 0, 0, m_game->m_width, m_game->m_height);
}

void
OpenGLDrawContext::draw_rectangle(float r, float g, float b, float a, int x, int y, int w, int h)
{
    drawTriangle(r, g, b, a, x, y, x+w, y, x, y+h);
    drawTriangle(r, g, b, a, x+w, y, x, y+h, x+w, y+h);
}

void
OpenGLDrawContext::draw_rectangle(enum Color color, int x, int y, int w, int h, float opacity)
{
    float r, g, b, a=opacity;
    RGBColor::map(color, r, g, b);
    draw_rectangle(r, g, b, a, x, y, w, h);
}

void
OpenGLDrawContext::draw_triangle(enum Color color, int x1, int y1, int x2, int y2, int x3, int y3, float opacity)
{
    float r, g, b, a=opacity;
    RGBColor::map(color, r, g, b);

    drawTriangle(r, g, b, a, x1, y1, x2, y2, x3, y3);
}

void
OpenGLDrawContext::draw_number(enum Color color, long number, int x, int y, int w, int h)
{
    int spacing = 10;
    draw_rectangle(0.f, 0.f, 0.f, 0.8f, x-spacing, y-spacing, w+spacing*2, h+spacing*2);

    setColor(color);

    GLocke::pushMatrix();
    GLocke::translate(x+w, y+h/2, 0.);
    GLocke::scale(h/2, -h/2, 1.);

    GLocke::translate(-.5, -.8, .2);
    GLocke::scale(3.2, 3.2, 1.);
    GLocke::rotate(20., 1., 0., 0.);
    drawFrameCount(number);

    GLocke::popMatrix();
}

void
OpenGLDrawContext::draw_text(enum Color color, const char *text, int x, int y, int w, int h, int flags)
{
    RenderedText rendered_text(text);

    GLocke::pushMatrix();

    if (flags & DrawContext::TEXT_CENTER) {
        x += (w - rendered_text.m_width*2) / 2.;
        y += (h - rendered_text.m_height*2) / 2.;
    }

    GLocke::translate(x, y, 0.);

    GLocke::scale(2., 2., 1.);

    setColor(color);
    rendered_text.render();

    GLocke::popMatrix();
}

void
OpenGLDrawContext::translate(int x, int y)
{
    GLocke::translate(x, y, 0.);
}

void
OpenGLDrawContext::draw_end()
{
    GLocke::end();
}

