/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/

#include "glocke.h"

#include <algorithm>
#include <stack>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#if defined(USE_OPENGL_ES)
#    include <GLES2/gl2.h>
#    define GLSL_PRECISION_INFO "precision mediump float;\n"
#elif defined(__APPLE__)
#    include <OpenGL/gl.h>
#else
#    define GL_GLEXT_PROTOTYPES
#    include <GL/gl.h>
#    include <GL/glext.h>
#endif

#if !defined(GLSL_PRECISION_INFO)
#define GLSL_PRECISION_INFO ""
#endif /* !defined(GLSL_PRECISION_INFO) */


const char *draw_colored_vertex_shader =
GLSL_PRECISION_INFO
"attribute vec4 vtxcoord;\n"
"attribute vec4 vtxcolor;\n"
"uniform mat4 projection;\n"
"varying vec4 col;\n"
"\n"
"void main()\n"
"{\n"
"    gl_Position = vtxcoord * projection;\n"
"    col = vtxcolor;\n"
"}\n"
;

const char *draw_colored_fragment_shader =
GLSL_PRECISION_INFO
"varying vec4 col;\n"
"\n"
"void main()\n"
"{\n"
"    gl_FragColor = col;\n"
"}\n"
;

const char *draw_alpha_textured_vertex_shader =
GLSL_PRECISION_INFO
"attribute vec4 vtxcoord;\n"
"attribute vec2 texcoord;\n"
"attribute vec4 vtxcolor;\n"
"uniform mat4 projection;\n"
"varying vec4 col;\n"
"varying vec2 tex;\n"
"\n"
"void main()\n"
"{\n"
"    gl_Position = vtxcoord * projection;\n"
"    col = vtxcolor;\n"
"    tex = texcoord;\n"
"}\n"
;

const char *draw_alpha_textured_fragment_shader =
GLSL_PRECISION_INFO
"uniform sampler2D sampler;\n"
"varying vec4 col;\n"
"varying vec2 tex;\n"
"\n"
"void main()\n"
"{\n"
"    gl_FragColor = vec4(col.rgb, texture2D(sampler, tex).a * col.a);\n"
"}\n"
;

const char *draw_lighting_vertex_shader =
GLSL_PRECISION_INFO
"attribute vec4 vtxcoord;\n"
"attribute vec3 nrmalvec;\n"
"uniform mat4 projection;\n"
"uniform mat4 modelview;\n"
"uniform vec4 color;\n"
"uniform vec3 lightpos;\n"
"varying vec4 col;\n"
"void main()\n"
"{\n"
"    gl_Position = vtxcoord * modelview * projection;\n"
"    vec3 p = vec3(vtxcoord * modelview);\n"
"    vec3 nn = normalize(vec3(vec4(nrmalvec, 0.0) * modelview));\n"
"    vec3 L = normalize(p - lightpos);\n"
"    float diffuse = clamp(dot(nn, L), 0.0, 1.0);\n"
"    col = vec4(color.rgb * (0.6 + 0.4 * diffuse), color.a);\n"
"}\n"
;

const char *draw_lighting_fragment_shader =
GLSL_PRECISION_INFO
"varying vec4 col;\n"
"void main()\n"
"{\n"
"    gl_FragColor = col;\n"
"}\n"
;


GLuint
make_program(const char *vertex_shader, const char *fragment_shader)
{
    GLuint program = glCreateProgram();

    GLuint vertex = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vertex, 1, &vertex_shader, 0);
    glCompileShader(vertex);
    glAttachShader(program, vertex);

    GLuint fragment = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(fragment, 1, &fragment_shader, 0);
    glCompileShader(fragment);
    glAttachShader(program, fragment);

    glLinkProgram(program);

    GLint result = GL_FALSE;
    glGetProgramiv(program, GL_LINK_STATUS, &result);
    if (!result) {
        printf("Failed to link program!\n");
        exit(0);
    }

    return program;
}






static std::stack<qwMatrix> modelview_stack;
static qwMatrix projection;


// Maximum numbers of attributes and uniforms
static const int GLES2_PROGRAM_MAX_ATTRIBUTES = 10;
static const int GLES2_PROGRAM_MAX_UNIFORMS = 10;

struct gles2_program_attrib_t {
    const char *name; // name of this attribute in the shader source
    int size; // number of floats for this attribute
    GLint location; // set to -1, will be filled by gles2_program_t::build
};

struct gles2_program_uniform_t {
    const char *name; // name of this uniform in the shader source
    GLint location; // set to -1, will be filled by gles2_program_t::build
};

struct gles2_program_t {
    void build() {
        program = make_program(vertex_shader_src, fragment_shader_src);
        stride = 0;

        gles2_program_attrib_t *attr = attributes;
        while (attr && attr->name) {
            attr->location = glGetAttribLocation(program, attr->name);
            stride += attr->size * sizeof(float);
            attr++;
        }

        gles2_program_uniform_t *unif = uniforms;
        while (unif && unif->name) {
            unif->location = glGetUniformLocation(program, unif->name);
            unif++;
        }
    }

    void enable() {
        glUseProgram(program);
        gles2_program_attrib_t *attr = attributes;
        size_t offset = 0;
        float *fbuffer = 0;
        while (attr && attr->name) {
            glEnableVertexAttribArray(attr->location);
            glVertexAttribPointer(attr->location, attr->size,
                    GL_FLOAT, GL_FALSE, stride, fbuffer + offset);
            offset += attr->size;
            attr++;
        }
    }

    void disable() {
        gles2_program_attrib_t *attr = attributes;
        while (attr && attr->name) {
            glDisableVertexAttribArray(attr->location);
            attr++;
        }
        glUseProgram(0);
    }

    const char *vertex_shader_src;
    const char *fragment_shader_src;

    gles2_program_attrib_t attributes[GLES2_PROGRAM_MAX_ATTRIBUTES];
    gles2_program_uniform_t uniforms[GLES2_PROGRAM_MAX_UNIFORMS];

    GLuint program; // generated by gles2_program_t::build
    size_t stride; // calculated from attributes by gles2_program_t::build, set to 0
};

static gles2_program_t draw_colored_shader_program = {
    draw_colored_vertex_shader, draw_colored_fragment_shader,
    {
        // Attributes:
        { "vtxcoord", 3 },
        { "vtxcolor", 4 },
        { NULL }
    },
    {
        // Uniforms:
        { "projection" },
        { NULL }
    },
};

static gles2_program_t draw_alpha_textured_shader_program = {
    draw_alpha_textured_vertex_shader, draw_alpha_textured_fragment_shader,
    {
        // Attributes:
        { "vtxcoord", 3 },
        { "texcoord", 2 },
        { "vtxcolor", 4 },
        { NULL }
    },
    {
        // Uniforms:
        { "projection" },
        { NULL }
    },
};

static gles2_program_t draw_lighting_shader_program = {
    draw_lighting_vertex_shader, draw_lighting_fragment_shader,
    {
        // Attributes:
        { "vtxcoord", 3 },
        { "nrmalvec", 3 },
        { NULL }
    },
    {
        // Uniforms:
        { "projection" },
        { "modelview" },
        { "color" },
        { "lightpos" },
        { NULL }
    },
};


struct accumulator_buffer_t {
    char *buffer;
    size_t offset;
    size_t length;

    accumulator_buffer_t()
        : buffer(0)
        , offset(0)
        , length(0)
    {
        resize(1024);
    }

    ~accumulator_buffer_t()
    {
        free(buffer);
    }

    void resize(size_t len)
    {
        length = len;
        buffer = (char *)realloc(buffer, len);
    }

    void append(void *data, size_t len)
    {
        if (offset + len > length) {
            resize(length * 2);
        }

        memcpy(buffer + offset, data, len);
        offset += len;
    }
};

class Accumulator {
public:
    Accumulator()
        : buffer()
    {
    }

    ~Accumulator()
    {
    }

    void submit(void *data, size_t len)
    {
        if (current != this) {
            if (current) {
                current->upload();
            }
            current = this;
        }

        buffer.append(data, len);
    }

    void upload()
    {
        if (buffer.offset) {
            render((float *)(buffer.buffer), buffer.offset);
            buffer.offset = 0;
        }
    }

    static void flush()
    {
        if (current) {
            current->upload();
            current = 0;
        }
    }

    virtual void render(float *buffer, size_t len) = 0;

private:
    accumulator_buffer_t buffer;
    static Accumulator *current;
};

Accumulator *
Accumulator::current = 0;

class Accumulator25D : public Accumulator {
public:
    virtual void render(float *buffer, size_t len);
};

class Accumulator2D : public Accumulator {
public:
    virtual void render(float *buffer, size_t len);
};

static Accumulator25D draw25d_accumulator;
static Accumulator2D draw2d_accumulator;

static GLuint vertex_buffer_mesh_data;
static GLuint vertex_buffer_stream;

struct glocke_color_t {
    float r;
    float g;
    float b;
    float a;

    operator float *() { return (float *)this; }
};

struct glocke_pos_t {
    float x;
    float y;
    float z;

    operator float *() { return (float *)this; }
};

static glocke_color_t current_color;
static glocke_pos_t light_pos;

struct GLockeAlphaTexturePriv {
    GLuint texture;
    int w;
    int h;
};

GLockeAlphaTexture::GLockeAlphaTexture(char *data, int w, int h)
    : priv(new GLockeAlphaTexturePriv)
{
    glGenTextures(1, &(priv->texture));
    priv->w = w;
    priv->h = h;

    glBindTexture(GL_TEXTURE_2D, priv->texture);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, w, h, 0, GL_ALPHA, GL_UNSIGNED_BYTE, data);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
}

GLockeAlphaTexture::~GLockeAlphaTexture()
{
    glDeleteTextures(1, &(priv->texture));
    delete priv;
}

void
GLocke::init()
{
    modelview_stack.push(qwMatrix());
    current_color.r = current_color.g = current_color.b = current_color.a = 0.f;

    glClearColor(0., 0., 0., 0.);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    draw_colored_shader_program.build();
    draw_alpha_textured_shader_program.build();
    draw_lighting_shader_program.build();

    glGenBuffers(1, &vertex_buffer_mesh_data);
    glGenBuffers(1, &vertex_buffer_stream);
}

bool
GLocke::resize(int &w, int &h)
{
    bool rotated = false;
    glViewport(0, 0, w, h);

    projection = qwMatrix();

    if (w > h) {
        std::swap(w, h);
        projection *= qwMatrix::rotation(90.f, 0.f, 0.f, 1.f);
        rotated = true;
    }

    projection *= qwMatrix::frustum(-w/2., w/2., h/2., -h/2., 10., 20.);
    projection *= qwMatrix::translation(-w/2., -h/2., -10.5);

    return rotated;
}

void
GLocke::begin()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    modelview_stack.top() = qwMatrix();
}

void
GLocke::clearDepth()
{
    glClear(GL_DEPTH_BUFFER_BIT);
}

void
GLocke::end()
{
    Accumulator::flush();
}

void
GLocke::draw(GLockeAlphaTexture *texture)
{
    Accumulator::flush();

    glDisable(GL_DEPTH_TEST);
    glBindTexture(GL_TEXTURE_2D, texture->priv->texture);

    glocke_color_t &c = current_color;

    float vertexptr[] = {
        0.f, 0.f, 0.f, 0.f, 0.f, c.r, c.g, c.b, c.a,
        0.f, texture->priv->h, 0.f, 0.f, 1.f, c.r, c.g, c.b, c.a,
        texture->priv->w, 0.f, 0.f, 1.f, 0.f, c.r, c.g, c.b, c.a,
        texture->priv->w, texture->priv->h, 0.f, 1.f, 1.f, c.r, c.g, c.b, c.a,
    };

    int stride = draw_alpha_textured_shader_program.stride;
    int elements = sizeof(vertexptr) / stride;

    applyModelView(3, stride, vertexptr, elements);

    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer_stream);
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertexptr), vertexptr, GL_STREAM_DRAW);

    draw_alpha_textured_shader_program.enable();
    glUniformMatrix4fv(draw_alpha_textured_shader_program.uniforms[0].location, 1, GL_FALSE, projection);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, elements);
    draw_alpha_textured_shader_program.disable();

    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void
GLocke::draw3d(int start, int count)
{
    Accumulator::flush();

    glEnable(GL_DEPTH_TEST);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer_mesh_data);
    draw_lighting_shader_program.enable();
    glUniformMatrix4fv(draw_lighting_shader_program.uniforms[0].location, 1, GL_FALSE, projection);
    glUniformMatrix4fv(draw_lighting_shader_program.uniforms[1].location, 1, GL_FALSE, modelviewMatrix());
    glUniform4fv(draw_lighting_shader_program.uniforms[2].location, 1, current_color);
    glDrawArrays(GL_TRIANGLES, start, count);
    draw_lighting_shader_program.disable();
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void
GLocke::draw2d(int vertices, float *vertexptr)
{
    draw2d_accumulator.submit(vertexptr, vertices * draw_colored_shader_program.stride);
}

void
Accumulator2D::render(float *buffer, size_t len)
{
    size_t vertices = len / draw_colored_shader_program.stride;

    glDisable(GL_DEPTH_TEST);

    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer_stream);
    glBufferData(GL_ARRAY_BUFFER, len, buffer, GL_STREAM_DRAW);

    draw_colored_shader_program.enable();
    glUniformMatrix4fv(draw_colored_shader_program.uniforms[0].location, 1, GL_FALSE, projection);
    glDrawArrays(GL_TRIANGLES, 0, vertices);
    draw_colored_shader_program.disable();

    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void
GLocke::draw25d(int vertices, float *vertexptr)
{
    draw25d_accumulator.submit(vertexptr, draw_colored_shader_program.stride * vertices);
}

void
Accumulator25D::render(float *buffer, size_t len)
{
    size_t vertices = len / draw_colored_shader_program.stride;

    glEnable(GL_DEPTH_TEST);

    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer_stream);
    glBufferData(GL_ARRAY_BUFFER, len, buffer, GL_STREAM_DRAW);

    draw_colored_shader_program.enable();
    glUniformMatrix4fv(draw_colored_shader_program.uniforms[0].location, 1, GL_FALSE, projection);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices);
    draw_colored_shader_program.disable();

    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void
GLocke::color(float r, float g, float b, float a)
{
    current_color.r = r;
    current_color.g = g;
    current_color.b = b;
    current_color.a = a;
}

void
GLocke::pushMatrix()
{
    modelview_stack.push(modelview_stack.top());
}

void
GLocke::popMatrix()
{
    modelview_stack.pop();
}

void
GLocke::translate(float x, float y, float z)
{
    modelview_stack.top() *= qwMatrix::translation(x, y, z);
}

void
GLocke::rotate(float angle, float x, float y, float z)
{
    modelview_stack.top() *= qwMatrix::rotation(angle, x, y, z);
}

void
GLocke::scale(float x, float y, float z)
{
    modelview_stack.top() *= qwMatrix::scaling(x, y, z);
}

void
GLocke::lightPosition(float x, float y, float z)
{
    light_pos.x = x;
    light_pos.y = y;
    light_pos.z = z;
}

void
GLocke::applyModelView(int size, int stride, float *vertexptr, int count)
{
    qwMatrix &m = modelview_stack.top();

    if (stride == 0) {
        // tightly packed
        stride = sizeof(float) * size;
    }

    for (int i=0; i<count; i++) {
        float *vtx = (float *)(((char *)vertexptr) + i * stride);
        float v[4];

        switch (size) {
            case 4:
                v[0] = vtx[0];
                v[1] = vtx[1];
                v[2] = vtx[2];
                v[3] = vtx[3];
                break;
            case 3:
                v[0] = vtx[0];
                v[1] = vtx[1];
                v[2] = vtx[2];
                v[3] = 1.f;
                break;
        }

        m.apply(v);

        switch (size) {
            default:
            case 4:
                vtx[0] = v[0];
                vtx[1] = v[1];
                vtx[2] = v[2];
                vtx[3] = v[3];
                break;
            case 3:
                vtx[0] = v[0] / v[3];
                vtx[1] = v[1] / v[3];
                vtx[2] = v[2] / v[3];
                break;
        }
    }
}

const qwMatrix &
GLocke::modelviewMatrix()
{
    return modelview_stack.top();
}

void
GLocke::uploadMeshData(void *data, size_t len)
{
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer_mesh_data);
    glBufferData(GL_ARRAY_BUFFER, len, data, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}
