/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/

#ifndef QW_QWTYPE_H
#define QW_QWTYPE_H

#include <math.h>
#include <string.h>

class qwMatrix {
    public:
        float m[4*4];

        qwMatrix()
        {
            for (int x=0; x<4; x++) {
                for (int y=0; y<4; y++) {
                    m[x + y*4] = (x==y)?1.:0.;
                }
            }
        }

        qwMatrix(float d0, float d1, float d2, float d3,
                 float d4, float d5, float d6, float d7,
                 float d8, float d9, float d10, float d11,
                 float d12, float d13, float d14, float d15)
        {
            float d[] = {
                d0, d1, d2, d3,
                d4, d5, d6, d7,
                d8, d9, d10, d11,
                d12, d13, d14, d15,
            };
            memcpy(m, d, sizeof(d));
        }

        operator const float *() const { return m; }

        qwMatrix transposed() const {
            return qwMatrix(
                    m[0], m[4], m[8], m[12],
                    m[1], m[5], m[9], m[13],
                    m[2], m[6], m[10], m[14],
                    m[3], m[7], m[11], m[15]
            );
        }

        void apply(float o[4]) const {
            float r[4];

            for (int i=0; i<4; i++) {
                r[i] = 0.f;
                for (int j=0; j<4; j++) {
                    r[i] += m[i*4 + j] * o[j];
                }
            }

            for (int i=0; i<4; i++) {
                o[i] = r[i];
            }
        }

        qwMatrix operator*(const qwMatrix &o) const {
            // http://www.gamedev.net/topic/379901-4x4-matrix-multiplication/
            qwMatrix result;

            result.m[0] = m[0] * o.m[0] + m[1] * o.m[4] + m[2] * o.m[8] + m[3] * o.m[12];
            result.m[1] = m[0] * o.m[1] + m[1] * o.m[5] + m[2] * o.m[9] + m[3] * o.m[13];
            result.m[2] = m[0] * o.m[2] + m[1] * o.m[6] + m[2] * o.m[10] + m[3] * o.m[14];
            result.m[3] = m[0] * o.m[3] + m[1] * o.m[7] + m[2] * o.m[11] + m[3] * o.m[15];

            result.m[4] = m[4] * o.m[0] + m[5] * o.m[4] + m[6] * o.m[8] + m[7] * o.m[12];
            result.m[5] = m[4] * o.m[1] + m[5] * o.m[5] + m[6] * o.m[9] + m[7] * o.m[13];
            result.m[6] = m[4] * o.m[2] + m[5] * o.m[6] + m[6] * o.m[10] + m[7] * o.m[14];
            result.m[7] = m[4] * o.m[3] + m[5] * o.m[7] + m[6] * o.m[11] + m[7] * o.m[15];

            result.m[8] = m[8] * o.m[0] + m[9] * o.m[4] + m[10] * o.m[8] + m[11] * o.m[12];
            result.m[9] = m[8] * o.m[1] + m[9] * o.m[5] + m[10] * o.m[9] + m[11] * o.m[13];
            result.m[10] = m[8] * o.m[2] + m[9] * o.m[6] + m[10] * o.m[10] + m[11] * o.m[14];
            result.m[11] = m[8] * o.m[3] + m[9] * o.m[7] + m[10] * o.m[11] + m[11] * o.m[15];

            result.m[12] = m[12] * o.m[0] + m[13] * o.m[4] + m[14] * o.m[8] + m[15] * o.m[12];
            result.m[13] = m[12] * o.m[1] + m[13] * o.m[5] + m[14] * o.m[9] + m[15] * o.m[13];
            result.m[14] = m[12] * o.m[2] + m[13] * o.m[6] + m[14] * o.m[10] + m[15] * o.m[14];
            result.m[15] = m[12] * o.m[3] + m[13] * o.m[7] + m[14] * o.m[11] + m[15] * o.m[15];

            return result;
        }

        void operator*=(const qwMatrix &o) {
            *this = *this * o;
        }

        static qwMatrix translation(float x, float y, float z) {
            // http://www.khronos.org/opengles/documentation/opengles1_0/html/glTranslate.html
            return qwMatrix(
                    1., 0., 0., x,
                    0., 1., 0., y,
                    0., 0., 1., z,
                    0., 0., 0., 1.
            );
        }

        static qwMatrix rotation(float angle, float x, float y, float z) {
            // http://www.khronos.org/opengles/documentation/opengles1_0/html/glRotate.html
            angle = angle * M_PI / 180.;
            float c = cosf(angle);
            float s = sinf(angle);

            // Normalize the (x, y, z) vector
            float l = sqrtf(x*x+y*y+z*z);
            x /= l;
            y /= l;
            z /= l;

            return qwMatrix(
                    x*x * (1.-c) + c, x*y * (1.-c) - z*s, x*z * (1.-c) + y*s, 0.,
                    x*y * (1.-c) + z*s, y*y * (1.-c) + c, y*z * (1.-c) - x*s, 0.,
                    x*z * (1.-c) - y*s, y*z * (1.-c) + x*s, z*z * (1.-c) + c, 0.,
                    0., 0., 0., 1.
            );
        }

        static qwMatrix scaling(float x, float y, float z) {
            // http://www.khronos.org/opengles/documentation/opengles1_0/html/glScale.html
            return qwMatrix(
                    x, 0., 0., 0.,
                    0., y, 0., 0.,
                    0., 0., z, 0.,
                    0., 0., 0., 1.
            );
        }

        static qwMatrix frustum(float left, float right, float bottom, float top,
                float near, float far) {
            // http://www.khronos.org/opengles/sdk/1.1/docs/man/glFrustum.xml
            float A = (right + left) / (right - left);
            float B = (top + bottom) / (top - bottom);
            float C = - (far + near) / (far - near);
            float D = - (2 * far * near) / (far - near);

            return qwMatrix(
                    2.f * near / (right - left), 0.f, A, 0.f,
                    0.f, 2.f * near / (top - bottom), B, 0.f,
                    0.f, 0.f, C, D,
                    0.f, 0.f, -1.f, 0.f
            );
        }
};

#endif /* QW_QWTYPE_H */
