/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "constants.h"
#include "drawcontext.h"
#include "game.h"

#include <ncurses.h>
#include <string.h>

#define SCALE_XX Constants::BRICK_WIDTH
#define SCALE_YY Constants::BRICK_HEIGHT

class TetrepeteteCurses : public DrawContext {
    public:
        TetrepeteteCurses();
        int main();
        ~TetrepeteteCurses();

        void draw_begin();
        void draw_brick(enum Color color, int x, int y, float rotation=0.f);
        void draw_empty(int x, int y, bool right, bool bottom);
        void draw_fading(float value, int x, int y);
        void draw_end();

        void draw_background(enum Color c1, enum Color c2);

        void draw_score(long ticks, long score, int x, int y, int w, int h);

        void draw_shade();
        void draw_rectangle(enum Color color, int x, int y, int w, int h, float opacity=1.f);
        void draw_triangle(enum Color color, int x1, int y1, int x2, int y2, int x3, int y3, float opacity=1.f);
        void draw_number(enum Color color, long number, int x, int y, int w, int h);
        void draw_title(int x, int y, int w, int h);

        void draw_text(enum Color color, const char *text, int x, int y, int w, int h, int flags);

        void translate(int x, int y);

    private:
        Game m_game;
        int m_x_offset;
        int m_y_offset;
};


int
main(int argc, char *argv[])
{
    TetrepeteteCurses tetrepetete;
    return tetrepetete.main();
}



TetrepeteteCurses::TetrepeteteCurses()
    : m_game(),
      m_x_offset(0),
      m_y_offset(0)
{
    initscr();
    raw();
    keypad(stdscr, TRUE);
    timeout(10);
    start_color();
    curs_set(0);
    noecho();

    init_pair(BLACK, COLOR_BLACK, COLOR_BLACK);
    init_pair(WHITE, COLOR_BLACK, COLOR_WHITE);
    init_pair(RED, COLOR_BLACK, COLOR_RED);
    init_pair(GREEN, COLOR_BLACK, COLOR_GREEN);
    init_pair(BLUE, COLOR_BLACK, COLOR_BLUE);
    init_pair(GREY, COLOR_BLACK, COLOR_WHITE);
    init_pair(LIGHT_GREEN, COLOR_BLACK, COLOR_GREEN);
}

enum Game::Key
convertKey(int key)
{
    switch (key) {
        case KEY_LEFT: return Game::LEFT;
        case KEY_RIGHT: return Game::RIGHT;
        case KEY_UP: return Game::UP;
        case KEY_DOWN: return Game::DOWN;
        case ' ': return Game::SPACE;
        default: return Game::IGNORED;
    }
}

int
TetrepeteteCurses::main()
{
    int win_h, win_w;
    getmaxyx(stdscr, win_h, win_w);

    m_game.resize((win_w-1)*SCALE_XX, (win_h-1)*SCALE_YY);

    while (!m_game.finished()) {
        int ch = getch();
        if (ch == 27) {
            break;
        }

        enum Game::Key key = convertKey(ch);
        if (key != Game::IGNORED) {
            m_game.keyboard(true, false, key);
            m_game.keyboard(false, false, key);
        }
        m_game.render(this);
    }

    return 0;
}

TetrepeteteCurses::~TetrepeteteCurses()
{
    curs_set(1);
    endwin();
}

void
TetrepeteteCurses::draw_begin()
{
    m_x_offset = m_y_offset = 0;
    erase();
}

void
TetrepeteteCurses::draw_brick(enum Color color, int x, int y, float rotation)
{
    int w = Constants::BRICK_WIDTH;
    int h = Constants::BRICK_HEIGHT;

    int col = (x / w) * 2;
    int row = y / h;

    col += m_x_offset;
    row += m_y_offset;

    int win_h, win_w;
    getmaxyx(stdscr, win_h, win_w);
    if (row < 0 || col < 0 || row > win_h - 1 || col > win_w - 2) {
        return;
    }

    int ch = ' ';

    int r = (int)rotation % 360;

    if ((r > 0 && r <= 45) || (r > 180 && r <= 225)) {
        ch = '/';
    } else if ((r > 45 && r <= 90) || (r > 225 && r <= 270)) {
        ch = '|';
    } else if ((r > 90 && r <= 135) || (r > 270 && r <= 315)) {
        ch = '\\';
    } else if ((r > 135 && r <= 180) || (r > 315 && r <= 360)) {
        ch = '=';
    }

    attron(COLOR_PAIR(color));
    move(row, col);
    addch(ch);
    move(row, col + 1);
    addch(ch);
    attroff(COLOR_PAIR(color));
}

void
TetrepeteteCurses::draw_empty(int x, int y, bool right, bool bottom)
{
    draw_brick(BLACK, x, y, 45);
}

void
TetrepeteteCurses::draw_fading(float value, int x, int y)
{
    draw_brick(WHITE, x, y, 0);
}

void
TetrepeteteCurses::draw_end()
{
    refresh();
}


void
TetrepeteteCurses::draw_background(enum Color c1, enum Color c2)
{
}

void
TetrepeteteCurses::draw_score(long ticks, long score, int x, int y, int w, int h)
{
    draw_number(GREY, score, x, y, w, h);
}

void
TetrepeteteCurses::draw_shade()
{
}

void
TetrepeteteCurses::draw_rectangle(enum Color color, int x, int y, int w, int h, float opacity)
{
    x /= SCALE_XX;
    y /= SCALE_YY;
    w /= SCALE_XX;
    h /= SCALE_YY;

    x += m_x_offset;
    y += m_y_offset;

    int win_h, win_w;
    getmaxyx(stdscr, win_h, win_w);

    attron(COLOR_PAIR(color));
    for (int xx=x; xx<=x+w; xx++) {
        for (int yy=y; yy<=y+h; yy++) {
            if (yy >= 0 && xx >= 0 && yy <= win_h - 1 && xx <= win_w - 2) {
                move(yy, xx);
                addch('#');
            }
        }
    }
    attroff(COLOR_PAIR(color));
}

void
TetrepeteteCurses::draw_triangle(enum Color color, int x1, int y1, int x2, int y2, int x3, int y3, float opacity)
{
}

void
TetrepeteteCurses::draw_number(enum Color color, long number, int x, int y, int w, int h)
{
    char tmp[1000];
    sprintf(tmp, "%ld", number);
    draw_text(color, tmp, x, y, w, h, DrawContext::TEXT_CENTER);
}

void
TetrepeteteCurses::draw_title(int x, int y, int w, int h)
{
    draw_text(WHITE, "Tetrepetete", x, y, w, h, DrawContext::TEXT_CENTER);
}


void
TetrepeteteCurses::draw_text(enum Color color, const char *text, int x, int y, int w, int h, int flags)
{
    x /= SCALE_XX;
    y /= SCALE_YY;
    w /= SCALE_XX;
    h /= SCALE_YY;

    if (flags & DrawContext::TEXT_CENTER) {
        y += (h - 1) / 2;
        x += (w - strlen(text)) / 2;
    }

    x += m_x_offset;
    y += m_y_offset;

    int win_h, win_w;
    getmaxyx(stdscr, win_h, win_w);

    while (*text) {
        if (y >= 0 && x >= 0 && y <= win_h - 1 && x <= win_w - 2) {
            move(y, x);
            addch(*text);
        }

        text++;
        x++;
    }
}

void
TetrepeteteCurses::translate(int x, int y)
{
    m_x_offset += x / SCALE_XX;
    m_y_offset += y / SCALE_YY;
}

