/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "gldraw.h"
#include "game.h"

#include <SDL/SDL.h>
#include <emscripten/emscripten.h>

// Implemented in the HTML page
extern "C" int get_window_width();
extern "C" int get_window_height();

// Touch screen events from JavaScript
#define EMTOUCHO_PRESS 0
#define EMTOUCHO_MOVE 1
#define EMTOUCHO_RELEASE 2

// Define this if you want to use emtoucho.js
#define USE_EMTOUCHO

class PlatformSDLWindow : public OpenGLDrawContext
{
    public:
        PlatformSDLWindow(Game *game)
            : OpenGLDrawContext(game)
        {
        }

        enum Game::Key
        convertKey(int key)
        {
            switch (key) {
                case SDLK_LEFT: return Game::LEFT;
                case SDLK_RIGHT: return Game::RIGHT;
                case SDLK_UP: return Game::UP;
                case SDLK_DOWN: return Game::DOWN;
                case SDLK_SPACE: return Game::SPACE;
                default: return Game::IGNORED;
            }
        }

        void
        init()
        {
	    SDL_Init(SDL_INIT_VIDEO);

	    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 1);
            SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
            SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
            SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 2);
            SDL_Surface *surface = SDL_SetVideoMode(get_window_width(), get_window_height(), 0, SDL_OPENGL);

            initializeGL();
            resizeGL(surface->w, surface->h);
        }

        void
        step()
        {
            SDL_Event e;
            if (!m_game->finished()) {
                while (SDL_PollEvent(&e)) {
                    switch (e.type) {
                        case SDL_KEYDOWN:
                            m_game->keyboard(true, false, convertKey(e.key.keysym.sym));
                            break;

                        case SDL_KEYUP:
                            m_game->keyboard(false, false, convertKey(e.key.keysym.sym));
                            break;

#if !defined(USE_EMTOUCHO)
                        case SDL_MOUSEBUTTONDOWN:
                            {
                                int x = e.button.x, y = e.button.y;
                                mapCoords(x, y);
                                m_game->touch(true, x, y);
                            }
                            break;

                        case SDL_MOUSEMOTION:
                            {
                                if (e.motion.state) {
                                    int x = e.motion.x, y = e.motion.y;
                                    mapCoords(x, y);
                                    m_game->motion(x, y);
                                }
                            }
                            break;

                        case SDL_MOUSEBUTTONUP:
                            {
                                int x = e.button.x, y = e.button.y;
                                mapCoords(x, y);
                                m_game->touch(false, x, y);
                            }
                            break;
#endif
                        default:
                            // do nothing
                            break;
                    }
                }

#if defined(USE_EMTOUCHO)
                while (emscripten_asm_const_int("return EmToucho.eventQueue.length;") > 0) {
                    emscripten_asm_const_int("EmToucho.currentEvent = EmToucho.eventQueue.shift();");

                    int id = emscripten_asm_const_int("return EmToucho.currentEvent.id;");
                    int state = emscripten_asm_const_int("return EmToucho.currentEvent.state;");
                    int x = emscripten_asm_const_int("return EmToucho.currentEvent.x;");
                    int y = emscripten_asm_const_int("return EmToucho.currentEvent.y;");

                    mapCoords(x, y);

                    if (state == EMTOUCHO_PRESS) {
                        m_game->touch(true, x, y);
                    } else if (state == EMTOUCHO_MOVE) {
                        m_game->motion(x, y);
                    } else {
                        m_game->touch(false, x, y);
                    }
                }
#endif

                paintGL();

                SDL_GL_SwapBuffers();
            }
        }
};

static PlatformSDLWindow *window = nullptr;

static void
step_main_loop()
{
    window->step();
}

int
main(int argc, char *argv[])
{
    window = new PlatformSDLWindow(new Game());
    window->init();
    emscripten_set_main_loop(step_main_loop, 60, 0);
    return 0;
}

