/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "constants.h"
#include "drawcontext.h"
#include "game.h"
#include "color.h"

#include "index_html.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libsoup/soup.h>

class HTMLEventSource : public DrawContext
{
    public:
        HTMLEventSource(Game *game, SoupServer *server, SoupMessage *msg,
                SoupMessageBody *body)
            : m_game(game),
              m_server(server),
              m_msg(msg),
              m_body(body)
        {
        }

        void send(char *msg) {
            soup_message_body_append(m_body, SOUP_MEMORY_STATIC, "data:", 5);
            soup_message_body_append(m_body, SOUP_MEMORY_COPY, msg, strlen(msg));
            soup_message_body_append(m_body, SOUP_MEMORY_STATIC, "\n\n", 2);
        }

        gboolean render() {
            m_game->render(this);
            soup_server_unpause_message(m_server, m_msg);
            return TRUE;
        }

        /* DrawContext */
        void draw_begin();
        void draw_background(enum Color c1, enum Color c2);
        void draw_brick(enum Color color, int x, int y, float rotation=0.f);
        void draw_empty(int x, int y, bool right, bool bottom);
        void draw_fading(float value, int x, int y);
        void draw_score(long ticks, long score, int x, int y, int w, int h);
        void draw_title(int x, int y, int w, int h);
        void draw_shade();
        void draw_rectangle(enum Color color, int x, int y, int w, int h, float opacity=1.f);
        void draw_triangle(enum Color color, int x1, int y1, int x2, int y2, int x3, int y3, float opacity=1.f);
        void draw_number(enum Color color, long number, int x, int y, int w, int h);
        void draw_text(enum Color color, const char *text, int x, int y, int w, int h, int flags);
        void translate(int x, int y);
        void draw_end();

    private:
        Game *m_game;

        SoupServer *m_server;
        SoupMessage *m_msg;
        SoupMessageBody *m_body;
};

class TetrepeteteHTML
{
    public:
        TetrepeteteHTML();
        int run();
        ~TetrepeteteHTML();

        void press(int x, int y) {
            m_game.touch(true, x, y);
            m_pressed = true;
        }

        void motion(int x, int y) {
            if (m_pressed) {
                m_game.motion(x, y);
            }
        }

        void release(int x, int y) {
            m_pressed = false;
            m_game.touch(false, x, y);
        }

    private:
        Game m_game;
        bool m_pressed;

        SoupServer *m_server;

        friend void on_events_request(SoupServer *, SoupMessage *, const char *, GHashTable *, SoupClientContext *, gpointer);
};

int
main(int argc, char *argv[])
{
    g_type_init();

    TetrepeteteHTML tetrepetete;
    return tetrepetete.run();
}

gboolean
on_render_to_browser(gpointer user_data)
{
    return static_cast<HTMLEventSource*>(user_data)->render();
}

void
on_home_request(SoupServer *server, SoupMessage *msg, const char *path,
        GHashTable *query, SoupClientContext *client, gpointer user_data)
{
    TetrepeteteHTML *tetrepetete = static_cast<TetrepeteteHTML*>(user_data);

    soup_message_set_status(msg, 200);

    SoupMessageBody *body;
    g_object_get(msg, "response-body", &body, NULL);

    soup_message_body_append(body, SOUP_MEMORY_STATIC, platform_html5_index_html,
            platform_html5_index_html_len);
}

void
on_events_request(SoupServer *server, SoupMessage *msg, const char *path,
        GHashTable *query, SoupClientContext *client, gpointer user_data)
{
    TetrepeteteHTML *tetrepetete = static_cast<TetrepeteteHTML*>(user_data);
    soup_message_set_status(msg, 200);

    SoupMessageBody *body;
    g_object_get(msg, "response-body", &body, NULL);

    SoupMessageHeaders *headers;
    g_object_get(msg, "response-headers", &headers, NULL);

    soup_message_headers_set_encoding(headers, SOUP_ENCODING_CHUNKED);
    soup_message_headers_set_content_type(headers, "text/event-stream", NULL);

    HTMLEventSource *source = new HTMLEventSource(&(tetrepetete->m_game),
            server, msg, body);

    g_timeout_add(30, on_render_to_browser, source);
}

void
on_input_request(SoupServer *server, SoupMessage *msg, const char *path,
        GHashTable *query, SoupClientContext *client, gpointer user_data)
{
    TetrepeteteHTML *tetrepetete = static_cast<TetrepeteteHTML*>(user_data);

    soup_message_set_status(msg, 200);

    int x = atoi((const char*)g_hash_table_lookup(query, "x"));
    int y = atoi((const char*)g_hash_table_lookup(query, "y"));

    char *type = (char*)g_hash_table_lookup(query, "type");

    if (strcmp(type, "press") == 0) {
        tetrepetete->press(x, y);
    } else if (strcmp(type, "motion") == 0) {
        tetrepetete->motion(x, y);
    } else if (strcmp(type, "release") == 0) {
        tetrepetete->release(x, y);
    }
}

TetrepeteteHTML::TetrepeteteHTML()
    : m_game(),
      m_pressed(false),
      m_server(soup_server_new(NULL))
{
    m_game.resize(Constants::WINDOW_WIDTH, Constants::WINDOW_HEIGHT);

    printf("http://localhost:%d/\n", soup_server_get_port(m_server));

    soup_server_add_handler(m_server, "/", on_home_request, this, NULL);
    soup_server_add_handler(m_server, "/events", on_events_request, this, NULL);
    soup_server_add_handler(m_server, "/input", on_input_request, this, NULL);
}

int
TetrepeteteHTML::run()
{
    soup_server_run(m_server);
    return 0;
}

TetrepeteteHTML::~TetrepeteteHTML()
{
}

void
HTMLEventSource::draw_begin()
{
    send("draw_begin()");
}

void
HTMLEventSource::draw_background(enum Color c1, enum Color c2)
{
    char *tmp = g_strdup_printf("draw_background(%d,%d)", (int)c1, (int)c2);
    send(tmp);
    g_free(tmp);
}

void
HTMLEventSource::draw_brick(enum Color color, int x, int y, float rotation)
{
    int w = Constants::BRICK_WIDTH;
    int h = Constants::BRICK_HEIGHT;

    draw_rectangle(color, x, y, w, h, 1.);
}

void
HTMLEventSource::draw_empty(int x, int y, bool right, bool bottom)
{
}

void
HTMLEventSource::draw_fading(float value, int x, int y)
{
    int w = Constants::BRICK_WIDTH;
    int h = Constants::BRICK_HEIGHT;
    draw_rectangle(GREY, x, y, w, h, value);
}

void
HTMLEventSource::draw_shade()
{
    send("draw_shade()");
}

void
HTMLEventSource::draw_rectangle(enum Color color, int x, int y, int w, int h, float opacity)
{
    char *tmp = g_strdup_printf("draw_rectangle(%d,%d,%d,%d,%d,%.2f)",
            color, x, y, w, h, opacity);
    send(tmp);
    g_free(tmp);
}

void
HTMLEventSource::draw_triangle(enum Color color, int x1, int y1, int x2, int y2, int x3, int y3, float opacity)
{
    char *tmp = g_strdup_printf("draw_triangle(%d,%d,%d,%d,%d,%d,%d,%.2f)",
            color, x1, y1, x2, y2, x3, y3, opacity);
    send(tmp);
    g_free(tmp);
}

void
HTMLEventSource::draw_number(enum Color color, long number, int x, int y, int w, int h)
{
    char *tmp = g_strdup_printf("%ld", number);
    draw_text(color, tmp, x, y, w, h, DrawContext::TEXT_CENTER);
    g_free(tmp);
}

void
HTMLEventSource::draw_text(enum Color color, const char *text, int x, int y, int w, int h, int flags)
{
    char *tmp = g_strdup_printf("draw_text(%d,'%s',%d,%d,%d,%d,%d)",
            color, text, x, y, w, h, flags);
    send(tmp);
    g_free(tmp);
}

void
HTMLEventSource::translate(int x, int y)
{
    char *tmp = g_strdup_printf("translate(%d,%d)", x, y);
    send(tmp);
    g_free(tmp);
}

void
HTMLEventSource::draw_score(long ticks, long score, int x, int y, int w, int h)
{
    draw_rectangle(GREY, x, y, w, h);
    char tmp[50];
    sprintf(tmp, "%ld", score);
    draw_text(WHITE, tmp, x, y, w, h, 0);
}

void
HTMLEventSource::draw_title(int x, int y, int w, int h)
{
    draw_text(WHITE, "Tetrepetete", x, y, w, h, TEXT_CENTER);
}

void
HTMLEventSource::draw_end()
{
    send("draw_end()");
}
