/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "constants.h"
#include "drawcontext.h"
#include "game.h"
#include "color.h"

#include "brick.xpm"

#include <QtGui>

class PlatformQtWidget : public QWidget, public DrawContext
{
    public:
        PlatformQtWidget(Game *game)
            : QWidget(),
              m_game(game)
        {
            startTimer(10);

            enum Color color = COLOR_FIRST;
            QImage brick_img;
            do {
                brick_img = QImage(brick_xpm);
                tintImage(color, brick_img);
                m_bricks[color] = QPixmap::fromImage(brick_img);
                color = (enum Color)((int)color + 1);
            } while (color <= COLOR_LAST);
        }

        /* DrawContext */
        void draw_begin();
        void draw_background(enum Color c1, enum Color c2);
        void draw_brick(enum Color color, int x, int y, float rotation=0.f);
        void draw_empty(int x, int y, bool right, bool bottom);
        void draw_fading(float value, int x, int y);
        void draw_score(long ticks, long score, int x, int y, int w, int h);
        void draw_title(int x, int y, int w, int h);
        void draw_shade();
        void draw_rectangle(enum Color color, int x, int y, int w, int h, float opacity=1.f);
        void draw_triangle(enum Color color, int x1, int y1, int x2, int y2, int x3, int y3, float opacity=1.f);
        void draw_number(enum Color color, long number, int x, int y, int w, int h);
        void draw_text(enum Color color, const char *text, int x, int y, int w, int h, int flags);
        void translate(int x, int y);
        void draw_end();

        void
        timerEvent(QTimerEvent *event)
        {
            if (m_game->finished()) {
                QApplication::quit();
                return;
            }
            update();
        }

        void
        paintEvent(QPaintEvent *event)
        {
            QPainter painter(this);

            painter.setBrush(Qt::black);
            painter.setPen(Qt::black);
            painter.drawRect(rect());

            m_painter = &painter;
            m_game->render(this);
        }

        enum Game::Key
        convertKey(int key)
        {
            switch (key) {
                case Qt::Key_Left: return Game::LEFT;
                case Qt::Key_Right: return Game::RIGHT;
                case Qt::Key_Up: return Game::UP;
                case Qt::Key_Down: return Game::DOWN;
                case Qt::Key_Space: return Game::SPACE;
                default: return Game::IGNORED;
            }
        }

        void
        keyPressEvent(QKeyEvent *event)
        {
            m_game->keyboard(true, event->isAutoRepeat(), convertKey(event->key()));

            if (event->key() == Qt::Key_P) {
                if (m_game->paused()) {
                    m_game->resume();
                } else {
                    m_game->pause();
                }
            }
        }

        void
        keyReleaseEvent(QKeyEvent *event)
        {
            m_game->keyboard(false, event->isAutoRepeat(), convertKey(event->key()));
        }

        void
        mousePressEvent(QMouseEvent *event)
        {
            m_game->touch(true, event->x(), event->y());
        }

        void
        mouseMoveEvent(QMouseEvent *event)
        {
            m_game->motion(event->x(), event->y());
        }

        void
        mouseReleaseEvent(QMouseEvent *event)
        {
            m_game->touch(false, event->x(), event->y());
        }

    private:
        Game *m_game;
        QPainter *m_painter;

        /* Colored bricks */
        QMap<enum Color,QPixmap> m_bricks;
        void tintImage(enum Color color, QImage &brick_img);

        /* Private drawing functions */
        void setColor(enum Color color, float opacity=1.f);
        void setGrey(float value);
        void fillRect(int x, int y, int w, int h, float rotation=0.f);
        void drawBrick(enum Color color, int x, int y, float rotation=0.f);
};

class TetrepeteteQt
{
    public:
        TetrepeteteQt();
        int main();
        ~TetrepeteteQt();

    private:
        Game m_game;
        PlatformQtWidget *m_window;
};

int
main(int argc, char *argv[])
{
    QApplication::setGraphicsSystem("raster");
    QApplication app(argc, argv);

    TetrepeteteQt tetrepetete;

    return app.exec();
}

TetrepeteteQt::TetrepeteteQt()
    : m_game()
{
    m_game.resize(Constants::WINDOW_WIDTH, Constants::WINDOW_HEIGHT);

    m_window = new PlatformQtWidget(&m_game);
    m_window->setWindowTitle(Constants::WINDOW_NAME);
    m_window->resize(Constants::WINDOW_WIDTH,
            Constants::WINDOW_HEIGHT);
    m_window->show();
}

TetrepeteteQt::~TetrepeteteQt()
{
    delete m_window;
}


void
PlatformQtWidget::draw_begin()
{
    setColor(BLACK);
}

void
PlatformQtWidget::draw_background(enum Color c1, enum Color c2)
{
    setColor(c1, .1);
    m_painter->drawRect(rect());
}

void
PlatformQtWidget::draw_brick(enum Color color, int x, int y, float rotation)
{
    if (color == SHADOW) {
        int w = Constants::BRICK_WIDTH;
        int h = Constants::BRICK_HEIGHT;
        setColor(color);
        fillRect(x, y, w, h, rotation);
        return;
    }

    drawBrick(color, x, y, rotation);
}

void
PlatformQtWidget::draw_empty(int x, int y, bool right, bool bottom)
{
    int w = Constants::BRICK_WIDTH;
    int h = Constants::BRICK_HEIGHT;

    setColor(BLACK, .5);
    fillRect(x, y, w, h);

    if (!right || !bottom) {
        setColor(WHITE, .5);

        if (!right) {
            m_painter->drawLine(x+w, y, x+w, y+h);
        }

        if (!bottom) {
            m_painter->drawLine(x, y+h, x+w, y+h);
        }
    }
}

void
PlatformQtWidget::draw_fading(float value, int x, int y)
{
    int w = Constants::BRICK_WIDTH;
    int h = Constants::BRICK_HEIGHT;

    setGrey(value);
    fillRect(x, y, w, h);
}

void
PlatformQtWidget::draw_shade()
{
    m_painter->fillRect(rect(), QColor(0, 0, 0, 200));
}

void
PlatformQtWidget::draw_rectangle(enum Color color, int x, int y, int w, int h, float opacity)
{
    setColor(color, opacity);
    m_painter->drawRect(QRectF(x, y, w, h));
}

void
PlatformQtWidget::draw_triangle(enum Color color, int x1, int y1, int x2, int y2, int x3, int y3, float opacity)
{
    setColor(color, opacity);
    QPoint triangle[] = {
        QPoint(x1, y1),
        QPoint(x2, y2),
        QPoint(x3, y3),
    };
    m_painter->drawPolygon(triangle, 3);
}

void
PlatformQtWidget::draw_number(enum Color color, long number, int x, int y, int w, int h)
{
    setColor(color);
    m_painter->drawText(QRectF(x, y, w, h), Qt::AlignCenter, QString::number(number));
}

void
PlatformQtWidget::draw_text(enum Color color, const char *text, int x, int y, int w, int h, int flags)
{
    setColor(color);
    int qt_flags = 0;
    if (flags & DrawContext::TEXT_CENTER) {
        qt_flags |= Qt::AlignCenter;
    }
    m_painter->drawText(QRectF(x, y, w, h), qt_flags, text);
}

void
PlatformQtWidget::translate(int x, int y)
{
    m_painter->translate(x, y);
}

void
PlatformQtWidget::draw_score(long ticks, long score, int x, int y, int w, int h)
{
    setColor(GREY);
    m_painter->drawRect(QRectF(x, y, w, h));
    char tmp[50];
    sprintf(tmp, "%ld", score);
    draw_text(WHITE, tmp, x, y, w, h, 0);
}

void
PlatformQtWidget::draw_title(int x, int y, int w, int h)
{
    draw_text(WHITE, "Tetrepetete", x, y, w, h, TEXT_CENTER);
}

void
PlatformQtWidget::draw_end()
{
    m_painter = NULL;
}

void
PlatformQtWidget::tintImage(enum Color color, QImage &brick_img)
{
    int r, g, b;
    RGBColor::map(color, r, g, b);

    /* Tint the grayscale palette of the brick with the given color */
    brick_img = brick_img.convertToFormat(QImage::Format_Indexed8);
    for (int i=0; i<brick_img.colorCount(); i++) {
        QColor c(QColor(brick_img.color(i)).lighter());
        brick_img.setColor(i, qRgb(r*c.redF(), g*c.greenF(), b*c.blueF()));
    }
}

void
PlatformQtWidget::setColor(enum Color color, float opacity)
{
    int r, g, b, a=(int)(255*opacity);

    RGBColor::map(color, r, g, b);
    m_painter->setBrush(QColor(r, g, b, a));
    m_painter->setPen(QColor(r, g, b, a));
}

void
PlatformQtWidget::setGrey(float value)
{
    int x = 255*value;
    m_painter->setBrush(QColor(x, x, x));
    m_painter->setPen(QColor(x, x, x));
}

void
PlatformQtWidget::fillRect(int x, int y, int w, int h, float rotation)
{
    m_painter->save();
    m_painter->translate(x + w/2, y + h/2);
    m_painter->rotate(rotation);
    m_painter->drawRect(-w/2, -h/2, w, h);
    m_painter->restore();
}

void
PlatformQtWidget::drawBrick(enum Color color, int x, int y, float rotation)
{
    int w = Constants::BRICK_WIDTH;
    int h = Constants::BRICK_HEIGHT;

    m_painter->save();
    m_painter->translate(x + w/2, y + h/2);
    m_painter->rotate(rotation);
    m_painter->drawPixmap(-w/2, -h/2, m_bricks[color]);
    m_painter->restore();
}

