/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "gldraw.h"

#include "game.h"
#include "sdlswipe.h"

#include <SDL.h>

class PlatformSDLWindow : public OpenGLDrawContext
{
    public:
        PlatformSDLWindow(Game *game)
            : OpenGLDrawContext(game),
              m_gameplay_active(false)
        {
        }

        enum Game::Key
        convertKey(int key)
        {
            switch (key) {
                case SDLK_LEFT: return Game::LEFT;
                case SDLK_RIGHT: return Game::RIGHT;
                case SDLK_UP: return Game::UP;
                case SDLK_DOWN: return Game::DOWN;
                case SDLK_SPACE: return Game::SPACE;
                default: return Game::IGNORED;
            }
        }

        int
        mainloop()
        {
	    SDL_Init(SDL_INIT_VIDEO);

#ifdef USE_OPENGL_ES
	    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 2);
            SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
            SDL_Surface *surface = SDL_SetVideoMode(0, 0, 0, SDL_OPENGLES | SDL_FULLSCREEN);
#else
            /* Antialiasing / multi-sampling on Desktop OpenGL */
            SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
            SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 4);
            SDL_Surface *surface = SDL_SetVideoMode(480*.7, 800*.7, 0, SDL_OPENGL);
            SDL_WM_SetCaption("Tetrepetete", "Tetrepetete");
#endif

            initializeGL();
            resizeGL(surface->w, surface->h);

            SDL_Event e;
            while (!m_game->finished()) {
                while (SDL_PollEvent(&e)) {
                    if (e.type == SDL_QUIT) return 0;

                    switch (e.type) {
                        case SDL_ACTIVEEVENT:
                            {
                                if (e.active.state == SDL_APPACTIVE && e.active.gain == 0) {
                                    m_game->pause();
                                    while (1) {
                                        SDL_WaitEvent(&e);
                                        if (e.type == SDL_ACTIVEEVENT) {
                                            if (e.active.state == SDL_APPACTIVE &&
                                                    e.active.gain == 1) {
                                                break;
                                            }
                                        } else if (e.type == SDL_QUIT) {
                                            return 0;
                                        }
                                    }
                                    m_game->resume();
                                }
                            }
                        case SDL_KEYDOWN:
                            if (e.key.keysym.sym == SDLK_q) {
                                return 0;
                            }
                            m_game->keyboard(true, false, convertKey(e.key.keysym.sym));
                            break;

                        case SDL_KEYUP:
                            m_game->keyboard(false, false, convertKey(e.key.keysym.sym));
                            break;

                        case SDL_MOUSEBUTTONDOWN:
                            {
                                int x = e.button.x, y = e.button.y;
                                mapCoords(x, y);
                                m_game->touch(true, x, y);
                            }
                            break;

                        case SDL_MOUSEMOTION:
                            {
                                if (e.motion.state) {
                                    int x = e.motion.x, y = e.motion.y;
                                    mapCoords(x, y);
                                    m_game->motion(x, y);
                                }
                            }
                            break;

                        case SDL_MOUSEBUTTONUP:
                            {
                                int x = e.button.x, y = e.button.y;
                                mapCoords(x, y);
                                m_game->touch(false, x, y);
                            }
                            break;
                        }
                }
                SDL_Delay(10);
                paintGL();

                bool gameplay_active = m_game->gameplay_active();
                if (m_gameplay_active != gameplay_active) {
                    set_swipe_lock(m_width, m_height, gameplay_active);
                    m_gameplay_active = gameplay_active;
                }

                SDL_GL_SwapBuffers();
            }

            return 0;
        }

    private:
        bool m_gameplay_active;
};

int
main(int argc, char *argv[])
{
    Game game;
    PlatformSDLWindow window(&game);
    return window.mainloop();
}

