/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "gldraw.h"

#include "game.h"

#include <SDL.h>

class PlatformSDLWindow : public OpenGLDrawContext
{
    public:
        PlatformSDLWindow(Game *game)
            : OpenGLDrawContext(game)
        {
        }

        enum Game::Key
        convertKey(int key)
        {
            switch (key) {
                case SDLK_LEFT: return Game::LEFT;
                case SDLK_RIGHT: return Game::RIGHT;
                case SDLK_UP: return Game::UP;
                case SDLK_DOWN: return Game::DOWN;
                case SDLK_SPACE: return Game::SPACE;
                default: return Game::IGNORED;
            }
        }

        int
        mainloop()
        {
	    SDL_Init(SDL_INIT_VIDEO);

	    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 2);
	    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 0);
            SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
            SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
            SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 2);

            m_window = SDL_CreateWindow("Tetrepetete",
                    SDL_WINDOWPOS_CENTERED,
                    SDL_WINDOWPOS_CENTERED,
                    540, 960,
                    SDL_WINDOW_SHOWN | SDL_WINDOW_OPENGL | SDL_WINDOW_FULLSCREEN);

            m_gl_context = SDL_GL_CreateContext(m_window);

            initializeGL();

            int w, h;
            SDL_GetWindowSize(m_window, &w, &h);
            resizeGL(w, h);

            SDL_Event e;
            while (!m_game->finished()) {
                while (SDL_PollEvent(&e)) {
                    if (e.type == SDL_QUIT) return 0;

                    switch (e.type) {
                        case SDL_KEYDOWN:
                            if (e.key.keysym.sym == SDLK_q) {
                                return 0;
                            }
                            m_game->keyboard(true, false, convertKey(e.key.keysym.sym));
                            break;

                        case SDL_KEYUP:
                            m_game->keyboard(false, false, convertKey(e.key.keysym.sym));
                            break;

                        case SDL_WINDOWEVENT:
                            if (e.window.event == SDL_WINDOWEVENT_FOCUS_LOST) {
                                m_game->pause();

                                // Render paused state once
                                paintGL();
                                SDL_GL_SwapWindow(m_window);

                                while (true) {
                                    SDL_WaitEvent(&e);
                                    if (e.type == SDL_QUIT) {
                                        return 0;
                                    } else if (e.type == SDL_WINDOWEVENT &&
                                            e.window.event == SDL_WINDOWEVENT_FOCUS_GAINED) {
                                        break;
                                    }
                                }
                                m_game->resume();
                            }
                            break;

                        case SDL_MOUSEBUTTONDOWN:
                            {
                                int x = e.button.x, y = e.button.y;
                                mapCoords(x, y);
                                m_game->touch(true, x, y);
                            }
                            break;

                        case SDL_MOUSEMOTION:
                            {
                                if (e.motion.state) {
                                    int x = e.motion.x, y = e.motion.y;
                                    mapCoords(x, y);
                                    m_game->motion(x, y);
                                }
                            }
                            break;

                        case SDL_MOUSEBUTTONUP:
                            {
                                int x = e.button.x, y = e.button.y;
                                mapCoords(x, y);
                                m_game->touch(false, x, y);
                            }
                            break;

                        case SDL_FINGERDOWN:
                            {
                                int x = e.tfinger.x, y = e.tfinger.y;
                                mapCoords(x, y);
                                m_game->touch(true, x, y);
                            }
                            break;

                        case SDL_FINGERMOTION:
                            {
                                int x = e.tfinger.x, y = e.tfinger.y;
                                mapCoords(x, y);
                                m_game->motion(x, y);
                            }
                            break;

                        case SDL_FINGERUP:
                            {
                                int x = e.tfinger.x, y = e.tfinger.y;
                                mapCoords(x, y);
                                m_game->touch(false, x, y);
                            }
                            break;
                        }
                }
                SDL_Delay(10);
                paintGL();

                SDL_GL_SwapWindow(m_window);
            }

            SDL_GL_DeleteContext(m_gl_context);
            SDL_DestroyWindow(m_window);
            SDL_Quit();

            return 0;
        }

    private:
        SDL_Window *m_window;
        SDL_GLContext m_gl_context;
};

int
main(int argc, char *argv[])
{
    Game game;
    PlatformSDLWindow window(&game);
    return window.mainloop();
}

