/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "tetrepetete.h"

#include "board.h"
#include "constants.h"
#include "game.h"
#include "util.h"

#include <vector>

#include <stdlib.h>

#define RECORD_EVENT_DATA(x, y) \
    if (!m_playback) { \
        m_recording.add(x, y, m_ticks); \
    }

#define RECORD_EVENT(x) \
    if (!m_playback) { \
        m_recording.add(x, m_ticks); \
    }


Board::Board(Game *game)
    : Scene(game),
      m_speedup(1.0f),
      m_window_width(-1),
      m_window_height(-1),
      m_x_offset(0),
      m_y_offset(0),
      m_ticks(0),
      m_fall_ticks(Constants::SLOW_FALL_TICKS),
      m_fall_accumulator(0.),
      m_randomizer(),
      m_colorscheme(),
      m_score(),
      m_fields(new Field[Constants::BOARD_WIDTH*Constants::BOARD_HEIGHT]),
      m_current_pattern(NULL),
      m_falling_bricks(),

      m_start_touch(),
      m_start_touch_time(0),
      m_last_touch(),
      m_moved(false),
      m_last_drop(0),
      m_can_drop(true),

      m_playback(false),
      m_recording(),

      m_pause_button_y(0)

{
    new_pattern();
}

void
Board::playback(const char *filename)
{
    m_recording.load(filename);
    m_ticks = 0;
    m_playback = true;
}


void
Board::get_score(const char *filename, long *score, long *lines)
{
    Board board(NULL);
    board.playback(filename);
    board.m_recording.playback_step(&board, -1);

    if (score) {
        *score = board.m_score.m_score;
    }

    if (lines) {
        *lines = board.m_score.m_lines;
    }
}

void
Board::playback_event(RecordingEvent event)
{
    switch (event.m_type) {
        case RecordingEvent::FALL: fall(); break;
        case RecordingEvent::DROP: drop(); break;
        case RecordingEvent::MOVE: move(event.m_data); break;
        case RecordingEvent::ROTATE: rotate(event.m_data); break;
        case RecordingEvent::NEW_PATTERN: new_pattern(event.m_data); break;
        case RecordingEvent::QUIT: if (m_game) m_game->event(Game::POP_CURRENT_SCENE); break;
        default: tetrepetete_assert(0); break;
    }
}

void
Board::resize(int width, int height)
{
    m_window_width = width;
    m_window_height = height;
    int spacing = m_window_width * .05;
    m_x_offset = spacing;
    m_y_offset = -spacing + (m_window_height - (Constants::SPACING*2 +
                (Constants::BOARD_SPACING + Constants::BRICK_HEIGHT) * Constants::BOARD_HEIGHT));
}

Board::~Board()
{
    delete [] m_fields;

    if (m_current_pattern) {
        delete m_current_pattern;
    }
}

void
Board::rotate(int direction)
{
    if (m_current_pattern) {
        RECORD_EVENT_DATA(RecordingEvent::ROTATE, direction);
        m_current_pattern->rotate(direction);

        int i;
        int column, row;
        struct Field *field;
        for (i=0; i<4; i++) {
            m_current_pattern->xy(i, column, row);

            field = at(column, row);
            if (field && field->m_occupied) {
                m_current_pattern->rotate(-direction);
                return;
            }
        }
    }
}


struct Field *
Board::at(int x, int y)
{
    if (x < 0 || x >= Constants::BOARD_WIDTH) return NULL;
    if (y < 0 || y >= Constants::BOARD_HEIGHT) return NULL;

    return &m_fields[y*Constants::BOARD_WIDTH + x];
}

bool
Board::fall(bool is_drop)
{
    if (!is_drop) {
        RECORD_EVENT(RecordingEvent::FALL);
    }

    m_current_pattern->fall(1);

    if (check_collision(m_current_pattern)) {
        m_current_pattern->fall(-1);
        freeze();
        return true;
    }

    return false;
}

void
Board::drop()
{
    RECORD_EVENT(RecordingEvent::DROP);

    int i;
    int column, row;
    struct Field *field;

    do {
        for (i=0; i<4; i++) {
            m_current_pattern->xy(i, column, row);
            field = at(column, row);
            if (field) {
                field->m_fading = true;
                field->m_fading_value = 1.;
            }
        }
        fade();
    } while (!fall(true));
}

void
Board::new_pattern(int id)
{
    if (m_current_pattern) {
        delete m_current_pattern;
    }

    if (id == -1) {
        id = m_randomizer.next();
    }

    m_current_pattern = new Pattern(id);

    RECORD_EVENT_DATA(RecordingEvent::NEW_PATTERN, id);
}

void
Board::tick()
{
    m_ticks++;

    if (m_playback) {
        m_recording.playback_step(this, 4);
    } else {
        m_fall_accumulator += 1.;
        float required_ticks = (float)m_fall_ticks / m_speedup;
        while (m_fall_accumulator > required_ticks) {
            fall();
            m_fall_accumulator -= required_ticks;
        }
    }

    int x, y;
    int i=m_falling_bricks.size()-1;

    while (i >= 0) {
        Brick *brick = m_falling_bricks[i];
        brick->xy(x, y);

        if ((y > m_window_height) ||
                ((x+Constants::BRICK_WIDTH) < 0) ||
                (x > m_window_width)) {
            delete brick;
            m_falling_bricks.erase(m_falling_bricks.begin() + i);
        } else {
            brick->step();
        }

        i--;
    }

    i = Constants::BOARD_WIDTH*Constants::BOARD_HEIGHT-1;
    while (i > 0) {
        if (m_fields[i].m_y_offset > 0) {
            m_fields[i].m_y_offset /= 2;
        }

        i--;
    }

    fade();
}

void
Board::transition(float value)
{
    m_pause_button_y = -Constants::PAUSE_BUTTON_SIZE * (1. - value);
}

void
Board::fade()
{
    int i;

    for (i=0; i<Constants::BOARD_WIDTH*Constants::BOARD_HEIGHT; i++) {
        struct Field *field = &(m_fields[i]);
        if (field->m_fading) {
            field->m_fading_value *= .85;
            if (field->m_fading_value < .1) {
                field->m_fading = false;
            }
        }
    }

}

void
Board::freeze()
{
    int i;
    int column, row;
    struct Field *field;

    if (m_current_pattern) {
        for (i=0; i<4; i++) {
            m_current_pattern->xy(i, column, row);
            field = at(column, row);
            if (field == NULL) {
                /* GAME OVER */
                if (m_game && !m_playback) {
                    m_recording.save(m_game->m_last_recording);
                    m_game->m_settings.m_last_score = m_score.m_score;
                    m_game->m_settings.m_last_lines = m_score.m_lines;
                    m_game->event(Game::BOARD_IS_FULL);
                }
                return;
            }
            field->m_pattern = m_current_pattern->pattern();
            field->m_occupied = true;
        }

        new_pattern();
    }

    remove_rows();
}

void
Board::remove_rows()
{
    struct Field *field;
    int i;
    int column, row;

    /**
     * This saves the negative Y-offset when multiple rows are removed
     * (for the falling bricks effect)
     **/
    int accumulator = 0;

    int lines_removed = 0;

    row = Constants::BOARD_HEIGHT-1;
    while (row >= 0) {
        bool full = true;
        for (column=0; column<Constants::BOARD_WIDTH; column++) {
            field = at(column, row);
            if (field && !field->m_occupied) {
                row--;
                full = false;
                accumulator = 0;
                break;
            }
        }

        if (!full) continue;

        lines_removed++;
        stop_fast();

        for (i=row; i>=0; i--) {
            for (column=0; column<Constants::BOARD_WIDTH; column++) {
                field = at(column, i);
                if (!field) continue;

                if (i > 0) {
                    if (i == row) {
                        int x, y;
                        brick_pos(column, row-accumulator, x, y);
                        m_falling_bricks.push_back(new Brick(field->m_pattern, x, y));
                    }

                    struct Field *f2 = at(column, (i-1));
                    if (f2) {
                        *field = *f2;
                        field->m_y_offset = Constants::BRICK_HEIGHT * (accumulator + 1);
                    }
                } else {
                    field->m_occupied = false;
                }
            }
        }

        accumulator++;
    }

    switch (lines_removed) {
        case 0: m_score.event(Game::PATTERN_FROZEN); break;
        case 1: m_score.event(Game::SINGLE_LINE_CLEAR); break;
        case 2: m_score.event(Game::DOUBLE_LINE_CLEAR); break;
        case 3: m_score.event(Game::TRIPLE_LINE_CLEAR); break;
        case 4: m_score.event(Game::PETETE_LINE_CLEAR); break;
        default: break;
    }

    int line_diff = (m_score.m_lines - m_score.m_lines_speedup_last);
    if (lines_removed && line_diff >= Constants::SPEEDUP_LINES) {
        m_score.m_lines_speedup_last = m_score.m_lines;
        m_colorscheme.next();
        m_speedup *= 1.1f;
    }
}

void
Board::move(int direction)
{
    if (m_current_pattern) {
        RECORD_EVENT_DATA(RecordingEvent::MOVE, direction);
        m_current_pattern->move(direction);

        int i;
        int column, row;
        struct Field *field;
        for (i=0; i<4; i++) {
            m_current_pattern->xy(i, column, row);
            if (column < 0 || column >= Constants::BOARD_WIDTH) {
                m_current_pattern->move(-direction);
                return;
            }

            field = at(column, row);
            if (field && field->m_occupied) {
                m_current_pattern->move(-direction);
                return;
            }
        }
    }
}

void
Board::render(DrawContext *context)
{
    int row, column;

    context->draw_background(m_colorscheme.map(-1), m_colorscheme.map(-2));

    /* Let the shadow pattern fall until it collides */
    Pattern shadow_pattern(*m_current_pattern);
    while (true) {
        shadow_pattern.fall(1);
        if (check_collision(&shadow_pattern)) {
            shadow_pattern.fall(-1);
            break;
        }
    }

    for (row=Constants::BOARD_HEIGHT-1; row>=0; row--) {
        for (column=0; column<Constants::BOARD_WIDTH; column++) {
            struct Field *field = at(column, row);
            int dx, dy;
            brick_pos(column, row, dx, dy);

            if (m_current_pattern->contains(column, row)) {
                context->draw_brick(m_colorscheme.map(m_current_pattern->pattern()), dx, dy);
            } else if (field->m_occupied) {
                context->draw_brick(m_colorscheme.map(field->m_pattern), dx, dy - field->m_y_offset);
            } else if (Constants::HAVE_SHADOW && shadow_pattern.contains(column, row)) {
                context->draw_brick(SHADOW, dx, dy);
            } else if (field->m_fading) {
                context->draw_fading(field->m_fading_value, dx, dy);
            } else {
                context->draw_empty(dx, dy,
                        column == Constants::BOARD_WIDTH-1,
                        row == Constants::BOARD_HEIGHT-1);
            }
        }
    }

    std::vector<Brick*>::iterator it;
    for (it=m_falling_bricks.begin(); it!=m_falling_bricks.end(); it++) {
        int pattern, x, y;

        (*it)->xy(x, y);
        pattern = (*it)->pattern();

        context->draw_brick(m_colorscheme.map(pattern), x, y, (*it)->rotation());
    }

    context->draw_score(m_ticks*33., m_score.m_score, m_window_width - 100, 20, 80, 80);

    context->draw_pause_button(m_pause_button_y, m_pause_button_y,
            Constants::PAUSE_BUTTON_SIZE, Constants::PAUSE_BUTTON_SIZE);
}

bool
Board::check_collision(Pattern *pattern)
{
    int i;
    int column, row;
    struct Field *field;
    for (i=0; i<4; i++) {
        pattern->xy(i, column, row);
        if (row >= Constants::BOARD_HEIGHT) {
            return true;
        }
        field = at(column, row);
        if (field && field->m_occupied) {
            return true;
        }
    }

    return false;
}


void
Board::brick_pos(int column, int row, int &x, int &y)
{
    x = m_x_offset + Constants::SPACING +
        column*(Constants::BRICK_WIDTH + Constants::BOARD_SPACING);
    y = m_y_offset + Constants::SPACING +
        row*(Constants::BRICK_HEIGHT + Constants::BOARD_SPACING);
}


void
Board::start_fast() {
    fall();
    m_fall_accumulator = 0;
    m_fall_ticks = Constants::FAST_FALL_TICKS;
}

void
Board::stop_fast() {
    m_fall_accumulator = 0;
    m_fall_ticks = Constants::SLOW_FALL_TICKS;
}

void
Board::touch(enum Scene::EventType event, int x, int y)
{
    if (event == Scene::UP) {
        if (x < Constants::PAUSE_BUTTON_SIZE && y < Constants::PAUSE_BUTTON_SIZE) {
            if (m_game) {
                m_game->event(Game::ASK_QUIT_BOARD);
            }
            return;
        }
    }

    if (m_playback) {
        /* Block input while playback is in progress */
        return;
    }

    if (event == Scene::DOWN) {
	m_start_touch.x = m_last_touch.x = x;
	m_start_touch.y = m_last_touch.y = y;
        m_start_touch_time = Util::now();
        m_moved = false;
	m_can_drop = true;
    } else if (event == Scene::UP) {
        float diff = (m_start_touch - Point(x, y)).length();

        if (!m_moved && diff < 20) {
            if (Constants::WANT_EASY_TOUCH) {
                int side_border = m_window_width / 6;
                int bottom_border = m_window_height / 8;
                if (y > m_window_height - bottom_border) {
                    drop();
                } else if (x < side_border) {
                    move(-1);
                } else if (x > m_window_width - side_border) {
                    move(1);
                } else {
                    rotate(1);
                }
            } else {
                rotate(1);
            }
        }
        m_last_drop = 0;
    } else {
        /* MOTION */
        Point cur(x, y);
        Point diff = cur - m_last_touch;

        long now = Util::now();

        float small_down = m_window_height / 26.f;
        float medium_sideways = m_window_width / 11.f;
        float large_down = m_window_height / 7.f;

        if (diff.y > small_down) {
            if (((now == m_last_drop) ||
                 (now - m_start_touch_time < 300 /* drop ms time */ &&
                  (cur - m_start_touch).length() > large_down)) &&
                    m_can_drop) {
                drop();
                m_last_drop = 0;
                m_can_drop = false;
            } else {
                fall();
                m_last_drop = now;
            }
            m_last_touch.x = cur.x;
            m_last_touch.y = cur.y;
            m_moved = true;
        } else if (diff.x > medium_sideways) {
            move(1);
            m_last_touch.x = cur.x;
            m_last_touch.y = cur.y;
            m_moved = true;
            m_can_drop = false;
        } else if (diff.x < -medium_sideways) {
            move(-1);
            m_last_touch.x = cur.x;
            m_last_touch.y = cur.y;
            m_moved = true;
            m_can_drop = false;
        }
    }
}

void
Board::keyboard(bool pressed, bool repeat, enum Game::Key key)
{
    if (m_playback) {
        /* Block input while playback is in progress */
        return;
    }

    if (pressed) {
        switch (key) {
            case Game::LEFT: move(-1); break;
            case Game::RIGHT: move(1); break;
            default: break;
        }
        if (!repeat) {
            switch (key) {
                case Game::UP: rotate(1); break;
                case Game::SPACE: drop(); break;
                case Game::DOWN: start_fast(); break;
                default: break;
            }
        }
    } else {
        if (!repeat) {
            switch (key) {
                case Game::DOWN: stop_fast(); break;
                default: break;
            }
        }
    }
}

void
Board::event(enum Game::Event event)
{
}

