/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "game.h"
#include "constants.h"
#include "board.h"
#include "drawcontext.h"
#include "util.h"
#include "transition.h"

#include "titlescene.h"
#include "confirmscene.h"
#include "menuscene.h"
#include "resultsscene.h"
#include "highscoresscene.h"

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

Game::Game(const char *data_dir)
    : m_settings(),
      m_scene_stack(),
      m_scene_garbage(),
      m_width(0),
      m_height(0),
      m_paused(false),
      m_last_time(Util::now()),
      m_tick_accumulator(0),
      m_transition(NULL),
      m_now(0),
      m_data_dir(NULL),
      m_last_recording(NULL),
      m_best_recording(NULL)
{
    push_scene(new TitleScene(this));
    srand((int)time(NULL));

    if (data_dir != NULL) {
        m_data_dir = strdup(data_dir);
    } else {
        m_data_dir = strdup(Util::get_data_dir());
    }

    Util::recursive_mkdir(m_data_dir);

    m_last_recording = Util::join_path(m_data_dir, "last.rec");
    m_best_recording = Util::join_path(m_data_dir, "best.rec");

    Board::get_score(m_best_recording,
            &(m_settings.m_best_score),
            &(m_settings.m_best_lines));
}

Game::~Game()
{
    while (!m_scene_stack.empty()) {
        delete m_scene_stack.top();
        m_scene_stack.pop();
    }

    if (m_transition) {
        delete m_transition;
    }

    free(m_data_dir);
    free(m_last_recording);
    free(m_best_recording);
}

void
Game::resize(int width, int height)
{
    m_width = width;
    m_height = height;

    scene()->resize(width, height);
}

void
Game::pause()
{
    if (gameplay_active()) {
        event(Game::ASK_QUIT_BOARD);
    }
    m_paused = true;
}

void
Game::resume()
{
    m_last_time = Util::now();
    m_paused = false;
}

long
Game::now()
{
    return m_now;
}

void
Game::render(DrawContext *context)
{
    if (m_transition && m_transition->done()) {
        delete m_transition;
        m_transition = NULL;
    }

    while (!m_transition && !m_scene_garbage.empty()) {
        delete m_scene_garbage.top();
        m_scene_garbage.pop();
    }

    if (!m_paused) {
        long now = Util::now();
        long diff = (now - m_last_time);
        m_now += diff;
        m_last_time = now;

        m_tick_accumulator += diff;
        while (m_tick_accumulator > Constants::TICK_MS) {
            if (m_transition) {
                m_transition->tick();
            } else {
                scene()->tick();
            }
            m_tick_accumulator -= Constants::TICK_MS;
        }
    }

    float partial_tick = (float)m_tick_accumulator / Constants::TICK_MS;

    context->draw_begin();
    context->draw_background(COLOR_PRIMARY, COLOR_SECONDARY);
    if (m_transition) {
        m_transition->render(context, partial_tick);
    } else {
        scene()->render(context);
    }
    context->draw_end();
}

bool
Game::finished()
{
    return m_scene_stack.empty();
}

bool
Game::gameplay_active()
{
    if (m_scene_stack.empty()) {
        return false;
    }

    return (m_scene_stack.top()->type() == Scene::GAMEPLAY);
}

void
Game::event(enum Game::Event event)
{
    switch (event) {
        case BOARD_IS_FULL:
             pop_scene();
             push_scene(new ResultsScene(this));
             break;

        case MAIN_MENU:
             push_scene(new MenuScene(this));
             break;

        case START_GAME:
             push_scene(new Board(this));
             break;

        case START_PLAYBACK:
             push_scene(new Board(this));
             ((Board*)scene())->playback(m_last_recording);
             break;

        case SHOW_HIGHSCORES:
             push_scene(new HighscoresScene(this));
             break;

        case ASK_QUIT_BOARD:
             push_scene(new ConfirmScene(this, scene(),
                         "Game Paused.",
                         "Resume", POP_CURRENT_SCENE,
                         "Quit", QUIT_BOARD));
             break;

        case QUIT_BOARD:
             pop_scene();
             pop_scene();
             break;

        case POP_CURRENT_SCENE:
             pop_scene();
             break;

        default: break;
    }

    // Dispatch event to current scene
    scene()->event(event);
}

void
Game::touch(bool pressed, int x, int y)
{
    if (pressed) {
        scene()->touch(Scene::DOWN, x, y);
    } else {
        scene()->touch(Scene::UP, x, y);
    }
}

void
Game::motion(int x, int y)
{
    scene()->touch(Scene::MOTION, x, y);
}

void
Game::keyboard(bool pressed, bool repeat, enum Key key)
{
    scene()->keyboard(pressed, repeat, key);
}

Scene *
Game::scene()
{
    if (!m_scene_stack.empty()) {
        return m_scene_stack.top();
    } else {
        return NULL;
    }
}

void
Game::push_scene(Scene *scene)
{
    if (!m_scene_stack.empty()) {
        if (scene->type() != Scene::OVERLAY) {
            if (m_transition) {
                delete m_transition;
            }
            m_transition = new Transition(this, m_scene_stack.top(), scene, true);
        }
    }

    scene->resize(m_width, m_height);
    m_scene_stack.push(scene);
    scene->event(SCENE_EVENT_PUSH);
}

void
Game::pop_scene()
{
    scene()->event(SCENE_EVENT_POP);

    Scene *from = m_scene_stack.top();
    /**
     * We can't delete the scene right now, as we might still need to keep it
     * alive for the transition, so just push it to the garbage stack. Once
     * the transition is over, all objects in the garbage stack will be freed.
     **/
    m_scene_garbage.push(from);
    m_scene_stack.pop();

    if (!m_scene_stack.empty()) {
        if (from->type() != Scene::OVERLAY) {
            if (m_transition) {
                delete m_transition;
            }
            m_transition = new Transition(this, from, scene(), false);
        }
    }

    scene()->resize(m_width, m_height);
    scene()->event(SCENE_EVENT_EXPOSE);
}

