/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/

#ifndef GAME_H
#define GAME_H

#include <math.h>
#include <stack>

#include "settings.h"

class Scene;
class DrawContext;
class Transition;

struct Point {
    int x;
    int y;

    Point(int x=0, int y=0) : x(x), y(y) {}

    Point
    operator-(Point other) {
        return Point(x - other.x, y - other.y);
    }

    float
    length() {
        return sqrtf(x*x + y*y);
    }
};

class Game {
    public:
        Game(const char *data_dir=0);
        ~Game();

        void resize(int width, int height);
        void render(DrawContext *context);
        void pause();
        void resume();

        bool paused() { return m_paused; }
        bool finished();

        bool gameplay_active();

        long now();

        enum Event {
            NONE = 0,

            PATTERN_FROZEN, /* = No line cleared */
            SINGLE_LINE_CLEAR,
            DOUBLE_LINE_CLEAR,
            TRIPLE_LINE_CLEAR,
            PETETE_LINE_CLEAR,

            BOARD_IS_FULL,

            MAIN_MENU,

            START_GAME,

            SHOW_HIGHSCORES,

            START_PLAYBACK,

            ASK_QUIT_BOARD,
            QUIT_BOARD,

            POP_CURRENT_SCENE,

            /* Events sent to the scene when pushed/popped */
            SCENE_EVENT_PUSH, // When scene is first pushed
            SCENE_EVENT_POP, // When scene is popped from the stack
            SCENE_EVENT_EXPOSE, // When scene is re-shown (scene on top is popped)
        };

        enum Key {
            IGNORED = 0,

            UP,
            DOWN,
            LEFT,
            RIGHT,

            SPACE,
        };

        void event(enum Event event);
        void touch(bool pressed, int x, int y);
        void motion(int x, int y);
        void keyboard(bool pressed, bool repeat, enum Key key);

    public:
        Settings m_settings;

    private:
        std::stack<Scene*> m_scene_stack;
        std::stack<Scene*> m_scene_garbage;

        Scene *
        scene();

        void
        push_scene(Scene *scene);

        void
        pop_scene();

    public:
        int m_width;
        int m_height;

    private:
        bool m_paused;
        long m_last_time;
        long m_tick_accumulator;

        Transition *m_transition;
        long m_now;

        char *m_data_dir;

    public:
        char *m_last_recording;
        char *m_best_recording;
};

#endif
