/**
 *  Tetrepetete - A lightweight and portable game with blocks.
 *  http://thp.io/2013/tetrepetete/
 *  Copyright (C) 2012, 2013, 2014, 2015 Thomas Perl <m@thp.io>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **/


#include "guiscene.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

struct GuiElement {
    public:
        GuiElement(GuiScene *scene, const char *text, int x, int y, int width, int height, enum Game::Event action);
        GuiElement(GuiScene *scene, const char *text, int x, int y, int width, int height, int flags, enum Color color);
        GuiElement(GuiScene *scene, bool *target, int x, int y, int width, int height);
        GuiElement(GuiScene *scene, long number, int x, int y, int width, int height);
        ~GuiElement();

        void
        touch(enum Scene::EventType event, int x, int y);

        void
        render(DrawContext *context);

        enum ElementType {
            BUTTON,
            BACK_BUTTON,
            NUMBER,
            TEXT,
        };

        GuiElement(GuiScene *scene, enum ElementType type, int x, int y, int width, int height);

        static enum Game::Event
        actionForType(enum ElementType type);

    private:
        GuiScene *m_scene;
        enum ElementType m_type;

        /* Text element */
        char *m_text;
        enum Color m_color;
        enum Game::Event m_action;

        /* Number element */
        long m_number;

        /* Generic attributes */
    public:
        int m_x;
        int m_y;
        int m_width;
        int m_height;
        bool m_visible;
        bool m_focused;

    private:
        /* Text element */
        int m_flags;
};


GuiElement::GuiElement(GuiScene *scene, const char *text, int x, int y,
        int width, int height, enum Game::Event action)
    : m_scene(scene),
      m_type(BUTTON),
      m_text(strdup(text)),
      m_color(WHITE),
      m_action(action),
      m_number(0),
      m_x(x),
      m_y(y),
      m_width(width),
      m_height(height),
      m_visible(true),
      m_focused(false),
      m_flags(0)
{
}

GuiElement::GuiElement(GuiScene *scene, long number, int x, int y, int width, int height)
    : m_scene(scene),
      m_type(NUMBER),
      m_text(NULL),
      m_color(WHITE),
      m_action(Game::NONE),
      m_number(number),
      m_x(x),
      m_y(y),
      m_width(width),
      m_height(height),
      m_visible(true),
      m_focused(false),
      m_flags(0)
{
}

GuiElement::GuiElement(GuiScene *scene, enum ElementType type, int x, int y, int width, int height)
    : m_scene(scene),
      m_type(type),
      m_text(NULL),
      m_color(WHITE),
      m_action(actionForType(type)),
      m_number(0),
      m_x(x),
      m_y(y),
      m_width(width),
      m_height(height),
      m_visible(true),
      m_focused(false),
      m_flags(0)
{
}

GuiElement::GuiElement(GuiScene *scene, const char *text, int x, int y,
        int width, int height, int flags, enum Color color)
    : m_scene(scene),
      m_type(TEXT),
      m_text(strdup(text)),
      m_color(color),
      m_action(Game::NONE),
      m_number(0),
      m_x(x),
      m_y(y),
      m_width(width),
      m_height(height),
      m_visible(true),
      m_focused(false),
      m_flags(flags)
{
}

GuiElement::~GuiElement()
{
    if (m_text) {
        free(m_text);
    }
}

enum Game::Event
GuiElement::actionForType(enum ElementType type)
{
    switch (type) {
        case BACK_BUTTON:
            return Game::POP_CURRENT_SCENE;
            break;
        default:
            return Game::NONE;
    }
}

void
GuiElement::touch(enum Scene::EventType event, int x, int y)
{
    if (event == Scene::UP) {
        if (!m_visible) return;
        if (x < m_x || x > m_x + m_width) return;
        if (y < m_y || y > m_y + m_height) return;

        switch (m_type) {
            case BUTTON:
            case BACK_BUTTON:
                m_scene->send_event(m_action);
                break;
            default:
                break;
        }
    }
}

void
GuiElement::render(DrawContext *context)
{
    if (!m_visible) return;

    switch (m_type) {
        case BUTTON:
            context->draw_button(COLOR_PRIMARY, m_text, m_x, m_y, m_width, m_height);
            break;
        case BACK_BUTTON:
            context->draw_back_button(m_x, m_y, m_width, m_height);
            break;
        case NUMBER:
            context->draw_number(WHITE, m_number, m_x, m_y, m_width, m_height);
            break;
        case TEXT:
            context->draw_text(m_color, m_text, m_x, m_y, m_width, m_height, m_flags);
            break;
    }

    if (m_focused) {
        context->draw_rectangle(WHITE, m_x, m_y, m_width, m_height, .5);
    }
}



GuiScene::GuiScene(Game *game)
    : Scene(game),
      m_elements(),
      m_back_button(NULL)
{
    m_back_button = new GuiElement(this, GuiElement::BACK_BUTTON,
                0, m_game->m_height,
                Constants::PAUSE_BUTTON_SIZE, Constants::PAUSE_BUTTON_SIZE);
    m_elements.push_back(m_back_button);
}

GuiScene::~GuiScene()
{
    for (int i=0; i<m_elements.size(); i++) {
        delete m_elements[i];
    }
}

void
GuiScene::touch(enum EventType event, int x, int y)
{
    for (int i=0; i<m_elements.size(); i++) {
        m_elements[i]->touch(event, x, y);
    }
}

void
GuiScene::keyboard(bool pressed, bool repeat, enum Game::Key key)
{
    if (!pressed) {
        int current = -1;
        for (int i=0; i<m_elements.size(); i++) {
            if (m_elements[i]->m_focused) {
                current = i;
                break;
            }
        }

        GuiElement *e = (current >= 0) ? m_elements[current] : NULL;

        if (key == Game::SPACE && e) {
            e->touch(Scene::DOWN, e->m_x, e->m_y);
            e->touch(Scene::UP, e->m_x, e->m_y);
        }

        if (key == Game::UP || key == Game::DOWN) {
            if (e) {
                e->m_focused = false;
                if (key == Game::UP) {
                    current--;
                } else if (key == Game::DOWN) {
                    current++;
                }

                if (current < 0) {
                    current = m_elements.size() -1;
                } else if (current > m_elements.size() - 1) {
                    current = 0;
                }
            } else {
                current = 0;
            }

            e = m_elements[current];
            e->m_focused = true;
        }
    }
}

void
GuiScene::render(DrawContext *context)
{
    for (int i=0; i<m_elements.size(); i++) {
        m_elements[i]->render(context);
    }
}

void
GuiScene::resize(int width, int height)
{
}

void
GuiScene::transition(float value)
{
    m_back_button->m_y = m_game->m_height - Constants::PAUSE_BUTTON_SIZE * value;
}

void
GuiScene::tick()
{
}

void
GuiScene::event(enum Game::Event event)
{
}

void
GuiScene::send_event(enum Game::Event event)
{
    m_game->event(event);
}

GuiElement *
GuiScene::add_button(const char *text, int x, int y, int width, int height, enum Game::Event action)
{
    GuiElement *result = new GuiElement(this, text, x, y, width, height, action);
    m_elements.push_back(result);
    return result;
}

GuiElement *
GuiScene::add_number(long number, int x, int y, int width, int height)
{
    GuiElement *result = new GuiElement(this, number, x, y, width, height);
    m_elements.push_back(result);
    return result;
}

GuiElement *
GuiScene::add_text(const char *text, int x, int y, int width, int height, int flags, enum Color color)
{
    GuiElement *result = new GuiElement(this, text, x, y, width, height, flags, color);
    m_elements.push_back(result);
    return result;
}

void
GuiScene::set_visible(GuiElement *element, bool visible)
{
    element->m_visible = visible;
}
